/* NEStra - The NES translator */
/* written by Quor */
/* Version 0.62 of May 22, 1999 */

/* Public Domain */


#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <stdio.h>

#include "mapper.h"
#include "io.h"
#include "globals.h"

u_char *ROM_BASE;
u_char *VROM_BASE;
unsigned int ROM_PAGES;
unsigned int VROM_PAGES;
int MAPPERNUMBER;
extern void(*MapperInit[])();
extern void translate();
char filename[1024];

int verbose=0;
int disassemble=0;
int dolink=0;
int mapmirror=0;
int oldxcode=0;
int irqflag=0;

void main(int argc,char **argv)
{
  int x;
  int r;
  int romfd;
  int size;
  int cmirror=0;

  /* Parse args */
  if (argc<2) {help();exit(1);}
  for (x=1;x<argc;x++)
  {
    if(strncmp(argv[x],"-v",3)==0) verbose=1;
    else if(strncmp(argv[x],"-d",2)==0) disassemble=1;
    else if(strncmp(argv[x],"-l",2)==0) dolink=1;
    else if(strncmp(argv[x],"-o",2)==0) oldxcode=1;
    else if(strncmp(argv[x],"-hm",3)==0) cmirror=1;
    else if(strncmp(argv[x],"-vm",3)==0) cmirror=2;
    else if(strncmp(argv[x],"-sm",3)==0) cmirror=3;
    else if(strncmp(argv[x],"-nm",3)==0) cmirror=4;
    else {strncpy(filename,argv[x],1020);break;}
  }
  
  /* Open ROM file */
  romfd=open(filename,O_RDONLY);
  if(romfd<0) {fprintf(stderr,"Unable to open %s\n",filename);fflush(stderr);exit(1);}
  size=lseek(romfd,0,SEEK_END);lseek(romfd,0,SEEK_SET); /* Get file size */
  if(size<0) {fprintf(stderr,"Unable to read %s\n",filename);fflush(stderr);exit(1);}
  if(size==0) {fprintf(stderr,"Unable to read %s (empty file)\n",filename);fflush(stderr);exit(1);}
  r=(int)mmap(ROM,0x1000000,PROT_READ,MAP_FIXED|MAP_PRIVATE,romfd,0);
  if(r<=0) {fprintf(stderr,"Unable to read %s\n",filename);fflush(stderr);exit(1);}

  /* Allocate memory */
  r=(int)mmap(RAM,0xC000,PROT_READ|PROT_WRITE,MAP_FIXED|MAP_PRIVATE|MAP_ANON,0,0);
  if(r<=0) {fprintf(stderr,"Out of memory!\n");fflush(stderr);exit(1);}
  r=(int)mmap(CODE_BASE,0x1000000,PROT_READ|PROT_WRITE|PROT_EXEC,MAP_FIXED|MAP_PRIVATE|MAP_ANON,0,0);
  if(r<=0) {fprintf(stderr,"Out of memory!\n");fflush(stderr);exit(1);}
  r=(int)mmap((void *)INT_MAP,0x300000,PROT_READ|PROT_WRITE,MAP_FIXED|MAP_PRIVATE|MAP_ANON,0,0);
  if(r<=0) {fprintf(stderr,"Out of memory!\n");fflush(stderr);exit(1);}

  /* Initialize ROM mapper */
  MAPTABLE[0]=RAM;
  MAPTABLE[1]=RAM;
  MAPTABLE[2]=RAM;
  MAPTABLE[4]=RAM;
  MAPTABLE[16]=RAM-65536; /* Wrap around accesses which index above FFFF */
  MAPTABLE[6]=MAPTABLE[7]=RAM; /* Save RAM */
  MAPPERNUMBER=-1;
  if(*((int *)ROM)==0x1a53454e) /* .NES */
  {
    ROM_BASE=ROM+16; /* 16 bytes for .nes header */
    MAPPERNUMBER=ROM[6]>>4;
    if(ROM[5]) memcpy(VRAM,ROM+16+ROM[4]*16384,8192);
    ROM_PAGES=ROM[4];
    VROM_PAGES=ROM[5];
    VROM_BASE=ROM+16+ROM[4]*16384;
    hvmirror=(ROM[6]&1)^1;
    nomirror=ROM[6]&8;
  }
  else if(size==40960)
  {
    /* No header, assume raw image */
    ROM_BASE=ROM;
    MAPPERNUMBER=0;
    memcpy(VRAM,ROM+32768,8192);
    ROM_PAGES=2;
    VROM_PAGES=2;
    VROM_BASE=ROM+32768;
    nomirror=1;
    mapmirror=0;
  }
  else {fprintf(stderr,"Unrecognized ROM file format\n");fflush(stderr);exit(1);}
  if(MAPPERNUMBER>MAXMAPPER||((int *)MapperInit)[MAPPERNUMBER]==0)
    {fprintf(stderr,"Unknown Mapper: %2x\n",MAPPERNUMBER);fflush(stderr);exit(1);}
  if(verbose) printf("Mapper %2x\n",MAPPERNUMBER);
  MapperInit[MAPPERNUMBER]();

  if(cmirror==1&&mapmirror==0)
  {
    if(verbose&&!hvmirror) printf("Overriding default vertical mirroring.\n");
    hvmirror=1;
  }
  if(cmirror==2&&mapmirror==0)
  {
    if(verbose&&hvmirror) printf("Overriding default horizontal mirroring.\n");
    hvmirror=0;
  }
  if(cmirror==3&&mapmirror==0)
  {
    if(verbose&&!nomirror&&!osmirror) printf("Overriding default mirroring.\n");
    osmirror=1;
  }
  if(cmirror==4&&mapmirror==0)
  {
    if(verbose&&!nomirror) printf("Overriding default mirroring.\n");
    nomirror=1;
  }
  if(verbose&&mapmirror==0)
  {
    if(nomirror) printf("Using no mirroring.\n");
    else if(osmirror) printf("Using one-screen mirroring.\n");
    else if(hvmirror) printf("Using horizontal mirroring\n");
    else printf("Using vertical mirroring\n");
  }

  restoresavedgame();

  /* Initialize graphic display */
  InitDisplay();

  /* start the show */
  STACKPTR=(int)STACK+0xFF;
  translate(*((unsigned short *)(MAPTABLE[15]+0xFFFC)));
  START(); /* execute translated code */

  /* Not Reached */
}

void quit()
{
  char buffer[1024];
  int fd;
  int x;
  /* Save ram */
  for(x=0;((long long *)(RAM+0x6000))[x]==0;x++) 
    if(x>=1023) exit(0); /* Nothing to save */
  strncpy(buffer,filename,1020);
  strcat(buffer,".sav");
  fd=open(buffer,O_CREAT|O_RDWR,0666);
  if(fd>0)
  {
    write(fd,RAM+0x6000,8192);
    close(fd);
    exit(0);
  }
  fprintf(stderr,"Could not write to save file %s\n",buffer);
  exit(1);
}  

restoresavedgame()
{
  char buffer[1024];
  int fd;
  strncpy(buffer,filename,1020);
  strcat(buffer,".sav");
  fd=open(buffer,O_RDWR,0666);
  read(fd,RAM+0x6000,8192);
  close(fd);
}  

help()
{
  fprintf(stderr,"Nestra v0.62\n\n");
  fprintf(stderr,"Usage: nestra [-v] [-d] [-l] filename\n");
  fprintf(stderr,"  -v Verbose output\n");
  fprintf(stderr,"  -d Disassemble\n");
  fprintf(stderr,"  -l Link branches optimization (may improve speed)\n");
  fprintf(stderr,"  -o Use old x11 display subroutines\n");
  fprintf(stderr,"  -hm Use horizontal mirroring\n");
  fprintf(stderr,"  -vm Use vertical mirroring\n");
  fprintf(stderr,"  -sm Use single-screen mirroring\n");
  fprintf(stderr,"  -nm Use no mirroring\n");
  fflush(stderr);
}
